# import permpy
from .permset import PermSet
from .permutation import Permutation
from .pegpermutation import PegPermutation
from .permclass import PermClass
from .avclass import AvClass
from .pegpermset import PegPermSet
from .geometricgridclass import GeometricGridClass
from .InsertionEncoding import *
import permpy.RestrictedContainer

import doctest
doctest.testmod()

